@@ -129,7 +129,7 @@ ONNX_ATTR_TO_REMOVE_PER_OP = {}
 
 TF_ATTR_TO_REMOVE = [
     "_output_shapes", "T", "seed2", "Tidx", "_class", "Tshape", "Tpaddings",
-    "data_format", "transpose_a", "transpose_b"
+    "data_format", "transpose_a", "transpose_b", "out_type"
 ]
 
 ONNX_OP_TO_TF_OP = {
@@ -173,14 +173,19 @@ TF_OP_STR_TO_ONNX_OP = {
     # these are temporarily added to
     # test other ops
     "Add": "Add",
+    "Ceil": "Ceil",
     "Equal": "Equal",
+    "Exp": "Exp",
+    "Floor": "Floor",
     "Greater": "Greater",
     "Identity": "Identity",
     "Less": "Less",
+    "Log": "Log",
     "LogicalAnd": "And",
     "LogicalNot": "Not",
     "LogicalOr": "Or",
     "LogicalXor": "Xor",
+    "LogSoftmax": "LogSoftmax",
     "MatMul": "MatMul",
     "Mul": "Mul",
     "Pow": "Pow",
@@ -257,3 +262,12 @@ def get_list_value(attr):
 
 def get_unique_suffix():
   return str(uuid.uuid4())[:8]
+
+
+def get_perm_from_formats(_from, _to):
+  return list(map(lambda x: _from.find(x), _to))
+
+
+# Constant string used to indicate that requested padding
+# is not natively supported in Tensorflow.
+PAD_TF_INCOMPATIBLE = "PAD_TF_INCOMPATIBLE"
